"""Top-level package for jsmetrics."""

__author__ = """Tom Keel"""
__email__ = "thomas.keel.18@ucl.ac.uk"
__version__ = "0.1.0"

from . import details_for_all_metrics
from .metrics import jet_metrics, jet_core_algorithms, waviness_metrics
